/*
 re:wind
  the Keyboard Launcher for Win32
   by Murachue
*/

/**************************************/
/* INCLUDES                           */
/**************************************/

#include "main.h"

/**************************************/
/* DEFINITIONS                        */
/**************************************/

static HWND hMainWnd;
static HWND hInputWnd;
static HWND hIconWnd;
static HWND hListContainerWnd;
static HWND hListWnd;
static HFONT hInputFont = NULL;
static HICON hFileIcon = NULL;
static UINT bgStatus;
static PATHLIST* rootpathlist = NULL;
static CMDLIST* rootcmdlist = NULL;

// setting.font.style
#define FONTSTYLE_BOLD 1
#define FONTSTYLE_ITALIC 2

static struct
{
	struct
	{
		char name[80];
		int size;
		DWORD style;
	}font;
	struct
	{
		COLORSET normal;
		COLORSET running;
		COLORSET notfound;
	}color;
	POINT lwndpos;
	struct
	{
		UINT modifier;
		UINT vkey;
	}hotkey;
} setting =
{
	"MS UI Gothic", 12, 0,
	RGB(220, 220, 255), RGB(0, 0, 0),
	RGB(255, 0, 0), RGB(255, 255, 255),
	RGB(255, 255, 255), RGB(192, 192, 192),
	100, 100,
	MOD_SHIFT | MOD_CONTROL, VK_SPACE
};

/**************************************/
/* MAIN CODES                         */
/**************************************/

char* getEnv(LPCTSTR name)
{
	char* tmp;
	DWORD size = 256;
	DWORD ret;

	for(;;)
	{
		tmp = halloc(size);
		ret = GetEnvironmentVariable(name, tmp, size);
		if(ret < size)
			break;	// yeah, I get it
		else
		{
			hfree(tmp);
			size += 256;
		}
	}

	return tmp;
}

char* getToken(char** p, char tok)
{
	char* token;
	char* boost;
	DWORD len;

	if(!**p)
		return 0;

	boost = *p;
	while(*boost && *boost != tok)
		boost++;
	len = boost - *p;
	if(token = halloc(len + 1))
	{
		strncopy(token, *p, len);
		token[len] = '\0';
	}
	*p = boost;
	if(**p == tok)
		(*p)++;

	return token;
}

DWORD getTokenDword(char** str, char tok)
{
	char* token, *p;
	DWORD val = 0;

	token = getToken(str, tok);
	for(p = token; '0' <= *p && *p <= '9'; p++)
	{
		val *= 10;
		val += *p - '0';
	}
	hfree(token);

	return val;
}

void setIcon(char* file)
{
	if(hFileIcon)
		DestroyIcon(hFileIcon);
	hFileIcon = NULL;
	if(!strlength(file))	// ȂɂȂ
		hFileIcon = LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_DEFICON), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
	if(!hFileIcon)	// system
	{
		CMDLIST* cl;

		for(cl = rootcmdlist; cl; cl = cl->next)
		{
			if(!strcomp(file, cl->name) && cl->icon)
			{
				char* fname, *clone = strclone(cl->icon), *p;
				DWORD iconno;

				p = clone;
				fname = getToken(&p, ',');
				iconno = getTokenDword(&p, ',');
				hfree(clone);
				// Hit!!
				//hFileIcon = LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_KILL), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
				if(strlength(fname))
					ExtractIconEx(fname, iconno, NULL, &hFileIcon, 1);
				else
				{
					char fn[MAX_PATH + 1];

					GetModuleFileName(NULL, fn, MAX_PATH);
					ExtractIconEx(fn, iconno, NULL, &hFileIcon, 1);
				}
				hfree(fname);
			}
		}
	}
	if(!hFileIcon)	// 𓀂Ă݂
		ExtractIconEx(file, 0, NULL, &hFileIcon, 1);
	if(!hFileIcon)	// .exeȂ̂?
	{
		char* tmpx;

		if(tmpx = halloc(strlength(file) + 5))
		{
			strcopy(tmpx, file);
			strplus(tmpx, ".exe");	// ܂exe
			ExtractIconEx(tmpx, 0, NULL, &hFileIcon, 1);
			hfree(tmpx);
		}
	}
	if(!hFileIcon)	// exeȂ֘AÂACRGet
	{
		SHFILEINFO sfi;	// ŏIiShell

		if(*(WORD*)file != '\\\\')	// ˂Ƃ킭fBNg΍
		{
			SHGetFileInfo(file, 0, &sfi, sizeof(sfi), SHGFI_ICON | SHGFI_SMALLICON);
			hFileIcon = sfi.hIcon;
		}
	}

	if(hFileIcon)
	{
		bgStatus = 0;
		SendMessage(hIconWnd, STM_SETICON, (WPARAM)hFileIcon, 0);
	}else
	{
		bgStatus = 2;
	}
	InvalidateRect(hMainWnd, NULL, TRUE);
	UpdateWindow(hMainWnd);

	return;
}

LRESULT CALLBACK ListboxContainerWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_SIZE:	// lParam̓NCAgTCYB
		SetWindowPos(hListWnd, NULL, 0, 0, LOWORD(lParam), HIWORD(lParam), SWP_NOZORDER | SWP_NOMOVE);
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

void complete(void)
{
	if(SendMessage(hListWnd, LB_GETCURSEL, 0, 0) != LB_ERR)
	{
		int sel = SendMessage(hListWnd, LB_GETCURSEL, 0, 0);
		char* str;
		char* in;

		if(!(str = halloc(SendMessage(hListWnd, LB_GETTEXTLEN, sel, 0) + 1)))
			return;	// ss\
		if(!(in = halloc(GetWindowTextLength(hInputWnd) + SendMessage(hListWnd, LB_GETTEXTLEN, sel, 0) + 1)))
			return;	// ss\
		GetWindowText(hInputWnd, in, GetWindowTextLength(hInputWnd) + 1);
		SendMessage(hListWnd, LB_GETTEXT, sel, (LPARAM)str);
		{
			char* p = basename(in);

			p = lastname(p, ':');

			strplus(in, str + strlength(p));
			SetWindowText(hInputWnd, in);
			SendMessage(hInputWnd, EM_SETSEL, strlength(in), strlength(in));
			SendMessage(hInputWnd, EM_SCROLLCARET, 0, 0);
		}

		if(*basename(in) != '\0')	// Not directory...tenuki
			ShowWindow(hListContainerWnd, SW_HIDE);

		hfree(in);
		hfree(str);
	}

	return;
}

LRESULT CALLBACK InputWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static WNDPROC preproc = NULL;

	switch(msg)
	{
	case WM_USER:
		preproc = (WNDPROC)GetWindowLong(hWnd, GWL_WNDPROC);
		return 0;
	case WM_CLOSE:
		break;
	case WM_KEYDOWN:
		switch(wParam)
		{
		case VK_RETURN:
			SendMessage((HWND)GetWindowLong(hWnd, GWL_HWNDPARENT), WM_USER, 0, 0);
			return 0;	// processed
		case VK_ESCAPE:
			SendMessage((HWND)GetWindowLong(hWnd, GWL_HWNDPARENT), WM_USER, 1, 0);
			return 0;	// processed
		case VK_DOWN:
			{
				LONG res;

				if((res = SendMessage(hListWnd, LB_GETCURSEL, 0, 0)) == LB_ERR)
					res = 0;
				else
					if(res == SendMessage(hListWnd, LB_GETCOUNT, 0, 0) - 1)
						res = 0;
					else
						res++;

				SendMessage(hListWnd, LB_SETCURSEL, res, 0);
			}
			return 0;	// processed
		case VK_UP:
			{
				LONG res;

				if((res = SendMessage(hListWnd, LB_GETCURSEL, 0, 0)) == LB_ERR)
					res = SendMessage(hListWnd, LB_GETCOUNT, 0, 0) - 1;
				else
					if(res == 0)
						res = SendMessage(hListWnd, LB_GETCOUNT, 0, 0) - 1;
					else
						res--;

				SendMessage(hListWnd, LB_SETCURSEL, res, 0);

				if(0)
				{
					DWORD len = GetWindowTextLength(hInputWnd);
					SendMessage(hMainWnd, WM_USER, 2, 0);
					complete();
					SendMessage(hInputWnd, EM_SETSEL, len, GetWindowTextLength(hInputWnd));
				}
			}
			return 0;	// processed
		case VK_TAB:
			complete();
			return 0;
		}
		break;
	case WM_CHAR:
		{
			// for beep issue
			if(wParam == 13)
				return 0;
			if(wParam == 27)
				return 0;
		}
		break;
	}

	return CallWindowProc(preproc, hWnd, msg, wParam, lParam);
}

void addEndYen(char* str)
{
	char* p = str;
	char* yen = str;

	while(*p)
	{
		if(!IsDBCSLeadByte(*p))
			if(*p == '\\')
				yen = p;
		p = CharNext(p);
	}

	if(yen != (str + strlength(str) - 1))
		strplus(str, "\\");

	return;
}

BOOL CALLBACK AboutDlgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_INITDIALOG:
		{
			char myname[MAX_PATH + 1];
			char* vi;
			DWORD size;

			GetModuleFileName(NULL, myname, MAX_PATH);
			size = GetFileVersionInfoSize(myname, 0);
			if(vi = halloc(size))
			{
				GetFileVersionInfo(myname, 0, size, vi);
				{
					VS_FIXEDFILEINFO* ver = (VS_FIXEDFILEINFO*)(vi + 40);

					// nameg
					wsprintf(myname, "%d.%d.%d.%d", HIWORD(ver->dwFileVersionMS), LOWORD(ver->dwFileVersionMS),
														HIWORD(ver->dwFileVersionLS), LOWORD(ver->dwFileVersionLS));
					SetDlgItemText(hWnd, IDC_VERSION, myname);
				}
				hfree(vi);
			}
		}
		return TRUE;	// set focus
	case WM_COMMAND:
		if(HIWORD(wParam) == 0)
			switch(LOWORD(wParam))
			{
			case IDOK:
				EndDialog(hWnd, 0);
				return TRUE;
			}
		break;
	}

	return FALSE;
}

void GetWindowPos(HWND hWnd, LPPOINT pos)
{
	RECT rc;

	GetWindowRect(hWnd, &rc);
	pos->x = rc.left;
	pos->y = rc.top;

	return;
}

void listResize(void)
{
	int max = SendMessage(hListWnd, LB_GETCOUNT, 0, 0);

	if(max == 0)	// XgȂ
		ShowWindow(hListContainerWnd, SW_HIDE);	// Xg̖
	else
	{
		// Select last one
		RECT rc;
		int height = (max > 5) ? 5 : max;

		SendMessage(hListWnd, LB_SETCURSEL, max - 1, 0);

		GetWindowRect(hMainWnd, &rc);
		SetWindowPos(hListContainerWnd, NULL, rc.left, rc.top - setting.font.size*height, 300, setting.font.size*height + 2, SWP_NOZORDER);
		ShowWindow(hListContainerWnd, SW_SHOWNOACTIVATE);
		SetFocus(hInputWnd);
	}

	return;
}

LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static HBRUSH hInputNormalBG;
	static HBRUSH hInputRunningBG;
	static HBRUSH hInputNotfoundBG;
	static BOOL dontcomplete = FALSE;

	switch(msg)
	{
	case WM_CREATE:
		// wtf
		//SetWindowLong(hWnd, GWL_STYLE, WS_POPUPWINDOW);
		//SetWindowLong(hWnd, GWL_EXSTYLE, WS_EX_TOOLWINDOW);
		hInputNormalBG = CreateSolidBrush(setting.color.normal.background);
		hInputRunningBG = CreateSolidBrush(setting.color.running.background);
		hInputNotfoundBG = CreateSolidBrush(setting.color.notfound.background);
		break;
	case WM_LBUTTONDOWN:
		SendMessage(hWnd, WM_NCLBUTTONDOWN, HTCAPTION, 0);
		break;
	case WM_NCLBUTTONDOWN:
		if(wParam == HTCAPTION)
			return DefWindowProc(hWnd, msg, wParam, lParam);
		else
			PostMessage(hWnd, WM_NCLBUTTONDOWN, HTCAPTION, lParam);
		break;
	case WM_COMMAND:
		switch(HIWORD(wParam))
		{
		case 0:	// normal command
			switch(LOWORD(wParam))
			{
			case ID_ABOUT:
				DialogBoxParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ABOUTDIALOG), hMainWnd, AboutDlgProc, 0);
				break;
			case ID_EXIT:
				DestroyWindow(hMainWnd);
				break;
			}
			break;
		case EN_CHANGE:	// ύX()܂
			{
				SIZE s;	// inputted string size
				char tmp[256];

				if(!dontcomplete)
				{
					// ⊮⏈
					// Xg
					if(!GetWindowTextLength(hInputWnd))	// ȂɂȂ
						ShowWindow(hListContainerWnd, SW_HIDE);	// XgȂ
					else
					{
						// ⊮
						PATHLIST* pl = rootpathlist;
						CMDLIST* cl = rootcmdlist;
						char* in;

						if(!(in = halloc(GetWindowTextLength(hInputWnd) + 1)))
							break;	// cannot continue
						GetWindowText(hInputWnd, in, GetWindowTextLength(hInputWnd) + 1);

						// PATH
						SendMessage(hListWnd, LB_RESETCONTENT, 0, 0);	// Xg{bNX
						while(pl)
						{
							WIN32_FIND_DATA wfd;
							HANDLE hf;

							strcopy(tmp, pl->dir);
							if(strlength(tmp))
								addEndYen(tmp);
							strplus(tmp, in);
							strplus(tmp, "*");
							if((hf = FindFirstFile(tmp, &wfd)) != INVALID_HANDLE_VALUE)
							{
								do
								{
									char* t;	// ܂Ƃ肠...߂͗ǂȂˁB

									if(!strcomp(wfd.cFileName, ".") || !strcomp(wfd.cFileName, ".."))
										continue;
									if(!(t = halloc(strlength(wfd.cFileName) + 2 + 4)))
										continue;	// Cannot continue
									strcopy(t, wfd.cFileName);
									if(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
										strplus(t, "\\");
									SendMessage(hListWnd, LB_INSERTSTRING, 0, (LPARAM)t);
									hfree(t);
								}while(FindNextFile(hf, &wfd));
								FindClose(hf);
							}

							pl = pl->next;
						}

						// R}hXg
						while(cl)
						{
							if(!strncomp(in, cl->name, GetWindowTextLength(hInputWnd)))
								SendMessage(hListWnd, LB_INSERTSTRING, 0, (LPARAM)cl->name);
							cl = cl->next;
						}

						hfree(in);

						// XgTCY
						listResize();
					}
				}else
				{
					dontcomplete = FALSE;
				}

				GetWindowText(hInputWnd, tmp, sizeof(tmp));
				{
					HDC hdc = GetDC(hInputWnd);
					HFONT hfold = SelectObject(hdc, hInputFont);
					GetTextExtentPoint32(hdc, tmp, strlength(tmp), &s);
					SelectObject(hdc, hfold);
					ReleaseDC(hInputWnd, hdc);
				}

				setIcon(tmp);

				// TCY
				SetWindowPos(hInputWnd, NULL, 0, 0, s.cx + 4 + 2, setting.font.size + 4, SWP_NOZORDER | SWP_NOMOVE);
				SetWindowPos(hMainWnd, NULL, 0, 0, s.cx + 4 + 2 + 18+1 + 4, ((setting.font.size + 4) > 18 ? setting.font.size + 4 : 18) + 4, SWP_NOZORDER | SWP_NOMOVE);

				// Edit̃r[: edv͕̂⊮őIĂꍇ邽..?
				{
					DWORD st, ed;

					// łł΋J͗vȂ
					//SendMessage(hInputWnd, WM_HSCROLL, SB_LEFT, (LPARAM)hInputWnd);

					SendMessage(hInputWnd, EM_GETSEL, (WPARAM)&st, (LPARAM)&ed);
					SendMessage(hInputWnd, EM_SETSEL, 0, 0);
					SendMessage(hInputWnd, EM_SCROLLCARET, 0, 0);
					SendMessage(hInputWnd, EM_SETSEL, st, ed);
				}
			}
			break;
		}
		break;
	case WM_CLOSE:
		SendMessage(hMainWnd, WM_USER, 1, 0);	// ESCƓ
		return 0;
	case WM_DESTROY:
		SendMessage(hInputWnd, WM_CLOSE, 0, 0);	// Edit, do settle.
		DeleteObject(hInputNormalBG);
		DeleteObject(hInputRunningBG);
		DeleteObject(hInputNotfoundBG);
		if(hFileIcon)
			DestroyIcon(hFileIcon);
		PostQuitMessage(0);
		break;
	case WM_HOTKEY:
		//MessageBox(NULL, "Hotkey", "ʼ_(OoO)^", MB_ICONINFORMATION | MB_OK);
		if(!(GetWindowLong(hMainWnd, GWL_STYLE) & WS_VISIBLE))
		{
			bgStatus = 0;
			SetWindowText(hInputWnd, "");
			ShowWindow(hMainWnd, SW_SHOW);
		}
		SetForegroundWindow(hMainWnd);
		SetFocus(hInputWnd);
		break;
	case WM_CTLCOLOREDIT:
		if((HWND)lParam == hInputWnd)
			switch(bgStatus)
			{
			case 1:
				SetBkColor((HDC)wParam, setting.color.running.background);
				SetTextColor((HDC)wParam, setting.color.running.foreground);
				return (LRESULT)hInputRunningBG;
			case 0:
				SetBkColor((HDC)wParam, setting.color.normal.background);
				SetTextColor((HDC)wParam, setting.color.normal.foreground);
				return (LRESULT)hInputNormalBG;
			case 2:
				SetBkColor((HDC)wParam, setting.color.notfound.background);
				SetTextColor((HDC)wParam, setting.color.notfound.foreground);
				return (LRESULT)hInputNotfoundBG;
			}
		break;
	case WM_USER:
		{
			switch(wParam)
			{
			case 0:	// ENTER: R U N !
				bgStatus = 1;
				InvalidateRect(hInputWnd, NULL, TRUE);
				UpdateWindow(hInputWnd);
				{
					char* in;

					if(!(in = halloc(GetWindowTextLength(hInputWnd) + 1)))
						break;	// cannot continue

					GetWindowText(hInputWnd, in, GetWindowTextLength(hInputWnd) + 1);
					{
						CMDLIST* cl;

						for(cl = rootcmdlist; cl; cl = cl->next)
							if(!strcomp(in, cl->name))
							{
								// ɏȂƂ?
								bgStatus = 0;
								ShowWindow(hListContainerWnd, SW_HIDE);
								SetWindowText(hInputWnd, "");

								switch(cl->type)
								{
								case CMDTYPE_SYS:
									switch(cl->param.syscmd)
									{
									case SYSCMD_EXIT:
										SendMessage(hMainWnd, WM_COMMAND, ID_EXIT, 0);
										break;
									case SYSCMD_ABOUT:
										SendMessage(hMainWnd, WM_COMMAND, ID_ABOUT, 0);
										break;
									default:
										MessageBox(hMainWnd, "sȃVXeR}hłB", "G[", MB_ICONWARNING | MB_OK);
									}
									break;
								case CMDTYPE_EXE:
									if((int)ShellExecute(hMainWnd, NULL, cl->param.execmd.exe, cl->param.execmd.param, cl->param.execmd.workdir, cl->param.execmd.nShowCmd) <= 32)
									{
										bgStatus = 2;
										InvalidateRect(hInputWnd, NULL, TRUE);
										UpdateWindow(hInputWnd);
										MessageBox(hMainWnd, cl->param.execmd.exe, "Can't execute", MB_ICONERROR | MB_OK);
									}else
									{
										ShowWindow(hMainWnd, SW_HIDE);
										ShowWindow(hListContainerWnd, SW_HIDE);
									}
									break;
								}
								hfree(in);
								break;
							}

						if(cl)	// executed.
							break;	// quit case
					}

					// None command, its (maybe) filename.
					{
						char* wkdir = strclone(in);

						*(basename(wkdir)) = '\0';
						if((int)ShellExecute(hMainWnd, NULL, in, NULL, wkdir, SW_SHOW) <= 32)
						{
							bgStatus = 2;
							InvalidateRect(hInputWnd, NULL, TRUE);
							UpdateWindow(hInputWnd);
						}else
						{
							ShowWindow(hMainWnd, SW_HIDE);
							ShowWindow(hListContainerWnd, SW_HIDE);
						}
						hfree(wkdir);
					}
					hfree(in);
				}
				break;
			case 1:	// ESCAPE
				if(GetWindowTextLength(hInputWnd))
				{
					SetWindowText(hInputWnd, "");
					bgStatus = 0;
				}else
				{
					ShowWindow(hListContainerWnd, SW_HIDE);
					ShowWindow(hMainWnd, SW_HIDE);
				}
				break;
			case 2:	// Dont Complete!!
				dontcomplete = TRUE;
				break;
			}
		}
		break;
	case WM_CONTEXTMENU:
		{
			HMENU hmparent = LoadMenu(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_LAUNCHERMENU));
			HMENU hm = GetSubMenu(hmparent, 0);

			TrackPopupMenu(hm, 0, LOWORD(lParam), HIWORD(lParam), 0, hMainWnd, NULL);

			DestroyMenu(hm);
			DestroyMenu(hmparent);
		}
		break;
	case WM_MOVE:
		GetWindowPos(hMainWnd, &setting.lwndpos);

		listResize();

		break;
	default:
		return DefWindowProc(hWnd, msg, wParam, lParam);
	}

	return 0;
}

int initWnd(void)
{
	WNDCLASSEX wc;
	HWND pre;

	ZeroMemory(&wc, sizeof(wc));
	wc.cbSize = sizeof(wc);
	wc.hbrBackground = (HBRUSH)GetStockObject(COLOR_BACKGROUND);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hInstance = GetModuleHandle(NULL);
	wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IDI_MAINICON));
	wc.lpfnWndProc = MainWndProc;
	wc.lpszClassName = "rewindLauncherWindowClass";
	if(!RegisterClassEx(&wc))
	{
		MessageBox(NULL, "RegisterClassExs", "_(OoO)^", MB_ICONERROR | MB_OK);
		return 1;
	}

	pre = GetForegroundWindow();
	if(!(hMainWnd = CreateWindowCX(WS_EX_TOOLWINDOW, wc.lpszClassName, "re:wind", WS_POPUPWINDOW,
									0, 0, 0, 0,
									NULL, NULL, wc.hInstance, NULL)))
	{
		MessageBox(NULL, "CreateWindow(Window)s", "_(OoO)^", MB_ICONERROR | MB_OK);
		return 2;
	}

	if(!(hInputWnd = CreateWindowEx(0, "EDIT", "", WS_CHILD | WS_VISIBLE | WS_BORDER | ES_AUTOHSCROLL,
									20, 1, 0, setting.font.size + 2,
									hMainWnd, NULL, wc.hInstance, NULL)))
	{
		MessageBox(NULL, "CreateWindow(Edit)s", "_(OoO)^", MB_ICONERROR | MB_OK);
		return 3;
	}

	SetWindowText(hInputWnd, "");	// AutoResize

	// Subclassing
	InputWndProc(hInputWnd, WM_USER, 0, 0);
	SetWindowLong(hInputWnd, GWL_WNDPROC, (LONG)InputWndProc);

	if(!(hIconWnd = CreateWindowEx(0, "STATIC", "", WS_CHILD | WS_VISIBLE | WS_BORDER | SS_ICON,
									1, 1, 18, 18,
									hMainWnd, NULL, wc.hInstance, NULL)))
	{
		MessageBox(NULL, "CreateWindow(static)s", "_(OoO)^", MB_ICONERROR | MB_OK);
		return 4;
	}

	//ZeroMemory(&wc, sizeof(wc));
	wc.cbSize = sizeof(wc);
	//wc.hbrBackground = (HBRUSH)GetStockObject(COLOR_BACKGROUND);
	//wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	//wc.hInstance = GetModuleHandle(NULL);
	//wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IDI_MAINICON));
	wc.lpfnWndProc = ListboxContainerWndProc;
	wc.lpszClassName = "rewindListboxContainerWC";
	if(!RegisterClassEx(&wc))
	{
		MessageBox(NULL, "RegisterClassEx(LBC)s", "_(OoO)^", MB_ICONERROR | MB_OK);
		return 1;
	}

	if(!(hListContainerWnd = CreateWindowEx(WS_EX_TOOLWINDOW, wc.lpszClassName, "", WS_POPUPWINDOW,
									0, 0, 0, 0,
									NULL, NULL, wc.hInstance, NULL)))
	{
		MessageBox(NULL, "CreateWindow(listboxcontainer)s", "_(OoO)^", MB_ICONERROR | MB_OK);
		return 5;
	}
	if(!(hListWnd = CreateWindowEx(0, "LISTBOX", "", WS_CHILD | WS_VISIBLE | WS_VSCROLL | LBS_NOTIFY,
									0, 0, 0, 0,
									hListContainerWnd, NULL, wc.hInstance, NULL)))
	{
		MessageBox(NULL, "CreateWindow(listbox)s", "_(OoO)^", MB_ICONERROR | MB_OK);
		return 6;
	}

	SetForegroundWindow(pre);

	return 0;
}

char* ReadLine(HANDLE hFile)
{
	char* str;
	DWORD index;
	DWORD size = 256;

	if(!(str = halloc(size)))
		return NULL;

	for(index = 0;;)
	{
		char c;
		DWORD read;

		if(!ReadFile(hFile, &c, 1, &read, NULL))	// error
			return NULL;
		if(read == 0)	// EOF
			if(index == 0)	// mou EOF dayo.
			{
				hfree(str);
				return NULL;
			}else
				break;

		if(c == 0x0D)	// next is 0x0A?
		{
			ReadFile(hFile, &c, 1, &read, NULL);
			if(c != 0x0A)	// oops! it is MAC format!! SHIT!!! F**K!!!!
				SetFilePointer(hFile, -1, NULL, FILE_CURRENT);
			break;
		}

		str[index++] = c;
		if(index + 2 >= size)	// NULL+1
		{
			size += 256;	// extend size
			if(!(str = hrealloc(str, size)))
				return NULL;	// failed.
		}
	}

	str[index] = '\0';

	return str;
}

BOOL isFileExist(char* fname)
{
	HANDLE hf;

	if((hf = CreateFile(fname, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL)) != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hf);
		return TRUE;
	}

	return FALSE;
}

void readConfig()
{
	char fn[MAX_PATH + 1];

	GetModuleFileName(NULL, fn, MAX_PATH);
	*basename(fn) = '\0';
	strplus(fn, "rewind.ini");

	if(isFileExist(fn))	// if ini exists, read.
	{
		GetPrivateProfileString("Settings", "font.name", setting.font.name, setting.font.name, sizeof(setting.font.name), fn);
		setting.font.size = GetPrivateProfileInt("Settings", "font.size", setting.font.size, fn);
		setting.font.style = GetPrivateProfileInt("Settings", "font.style", setting.font.style, fn);
		setting.color.normal.foreground = GetPrivateProfileInt("Settings", "color.normal.foreground", setting.color.normal.foreground, fn);
		setting.color.normal.background = GetPrivateProfileInt("Settings", "color.normal.background", setting.color.normal.background, fn);
		setting.color.running.foreground = GetPrivateProfileInt("Settings", "color.running.foreground", setting.color.running.foreground, fn);
		setting.color.running.background = GetPrivateProfileInt("Settings", "color.running.background", setting.color.running.background, fn);
		setting.color.notfound.foreground = GetPrivateProfileInt("Settings", "color.notfound.foreground", setting.color.notfound.foreground, fn);
		setting.color.notfound.background = GetPrivateProfileInt("Settings", "color.notfound.background", setting.color.notfound.background, fn);
		setting.lwndpos.x = GetPrivateProfileInt("Settings", "lwndpos.x", setting.lwndpos.x, fn);
		setting.lwndpos.y = GetPrivateProfileInt("Settings", "lwndpos.y", setting.lwndpos.y, fn);
		setting.hotkey.modifier = GetPrivateProfileInt("Settings", "hotkey.modifier", setting.hotkey.modifier, fn);
		setting.hotkey.vkey = GetPrivateProfileInt("Settings", "hotkey.vkey", setting.hotkey.vkey, fn);
	}

	return;
}

void WritePrivateProfileInt(LPCTSTR lpAppName, LPCTSTR lpKeyName, UINT nInt, LPCTSTR lpFileName)
{
	char tmp[11];	// 32bitUnsignedMax = 4294967296 = 10 + NULL

	wsprintf(tmp, "%u", nInt);
	WritePrivateProfileString(lpAppName, lpKeyName, tmp, lpFileName);

	return;
}

void saveConfig()
{
	char fn[MAX_PATH + 1];

	GetModuleFileName(NULL, fn, MAX_PATH);
	*basename(fn) = '\0';
	strplus(fn, "rewind.ini");

	WritePrivateProfileString("Settings", "font.name", setting.font.name, fn);
	WritePrivateProfileInt("Settings", "font.size", setting.font.size, fn);
	WritePrivateProfileInt("Settings", "font.style", setting.font.style, fn);
	WritePrivateProfileInt("Settings", "color.normal.foreground", setting.color.normal.foreground, fn);
	WritePrivateProfileInt("Settings", "color.normal.background", setting.color.normal.background, fn);
	WritePrivateProfileInt("Settings", "color.running.foreground", setting.color.running.foreground, fn);
	WritePrivateProfileInt("Settings", "color.running.background", setting.color.running.background, fn);
	WritePrivateProfileInt("Settings", "color.notfound.foreground", setting.color.notfound.foreground, fn);
	WritePrivateProfileInt("Settings", "color.notfound.background", setting.color.notfound.background, fn);
	WritePrivateProfileInt("Settings", "lwndpos.x", setting.lwndpos.x, fn);
	WritePrivateProfileInt("Settings", "lwndpos.y", setting.lwndpos.y, fn);
	WritePrivateProfileInt("Settings", "hotkey.modifier", setting.hotkey.modifier, fn);
	WritePrivateProfileInt("Settings", "hotkey.vkey", setting.hotkey.vkey, fn);

	return;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
	MSG msg;

	if(initWnd())
		ExitProcess(1);

	//GetWindowPos(hMainWnd, &setting.lwndpos);	// for no exist ini
	// ȂinitWnd>CreateWindowCX>(myrt.c)CreateWindowExŃj󂪋N̂őΏ
	setting.lwndpos.x = 100;
	setting.lwndpos.y = 100;
	readConfig();
	SetWindowPos(hMainWnd, NULL, setting.lwndpos.x, setting.lwndpos.y, 0, 0, SWP_NOZORDER | SWP_NOSIZE);

	// zbgL[
	if(!RegisterHotKey(hMainWnd, 0, MOD_SHIFT | MOD_CONTROL, VK_SPACE))
	{
		MessageBox(NULL, "RegisterHotKeys", "_(OoO)^", MB_ICONERROR | MB_OK);
		ExitProcess(2);
	}

	{
		LOGFONT lf;

		ZeroMemory(&lf, sizeof(LOGFONT));
		lf.lfHeight = setting.font.size;	// +px -pt
		lf.lfWeight = (setting.font.style & FONTSTYLE_BOLD) ? FW_BOLD : 0;
		lf.lfItalic = (setting.font.style & FONTSTYLE_ITALIC) ? TRUE : FALSE;
		lf.lfCharSet = DEFAULT_CHARSET;
		strcopy(lf.lfFaceName, setting.font.name);
		if(!(hInputFont = CreateFontIndirect(&lf)))
			MessageBox(NULL, "tHg쐬s", "G[", MB_ICONERROR | MB_OK);

		SendMessage(hInputWnd, WM_SETFONT, (WPARAM)hInputFont, TRUE);
		SendMessage(hListWnd, WM_SETFONT, (WPARAM)hInputFont, TRUE);
	}

	{
		char* path, *p;
		int i;
		PATHLIST* pl = NULL;

		path = getEnv("path");
		//MessageBox(hWnd, tmp, "env", MB_OK);
		//SendMessage(hListWnd, LB_RESETCONTENT, 0, 0);
		if(!(path = hrealloc(path, strlength(path) + 3)))
		{
			MessageBox(hMainWnd, "PATHgs", "G[", MB_ICONERROR | MB_OK);
			ExitProcess(1);
		}
		strplus(path, ";;");	// currentWbN
		for(p = path, i = 0;;)
		{
			char* token;

			if(!(token = getToken(&p, ';')))	// Error, or no more token.
				break;
			//SendMessage(hListWnd, LB_INSERTSTRING, i++, (LPARAM)token);
			if(!pl)
			{
				if(!(rootpathlist = pl = halloc(sizeof(PATHLIST))))
				{
					MessageBox(hMainWnd, "PATHi[pmێs", "G[", MB_ICONERROR | MB_OK);
					ExitProcess(1);
				}
			}else
			{
				if(!(pl->next = halloc(sizeof(PATHLIST))))
				{
					MessageBox(hMainWnd, "PATHi[pmێs", "G[", MB_ICONERROR | MB_OK);
					ExitProcess(1);
				}
				pl = pl->next;
			}
			if(!(pl->dir = halloc(strlength(token) + 1)))
			{
				MessageBox(hMainWnd, "PATHfBNgi[pmێs", "G[", MB_ICONERROR | MB_OK);
				ExitProcess(1);
			}
			strcopy(pl->dir, token);
			pl->next = NULL;
			hfree(token);
		}
		hfree(path);
	}

	{
		HANDLE hf;
		char fn[MAX_PATH + 1];

		GetModuleFileName(NULL, fn, MAX_PATH);
		*basename(fn) = '\0';
		strplus(fn, "cmdlist.txt");
		if((hf = CreateFile(fn, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) != INVALID_HANDLE_VALUE)
		{
			char* line;
			CMDLIST* cl = NULL;

			while(line = ReadLine(hf))
			{
				if(!cl)
				{
					if(!(rootcmdlist = cl = halloc(sizeof(CMDLIST))))
					{
						MessageBox(hMainWnd, "R}hXgi[pmێs", "G[", MB_ICONERROR | MB_OK);
						ExitProcess(1);
					}
				}else
				{
					if(!(cl->next = halloc(sizeof(CMDLIST))))
					{
						MessageBox(hMainWnd, "R}hXgi[pmێs", "G[", MB_ICONERROR | MB_OK);
						ExitProcess(1);
					}
					cl = cl->next;
				}

				ZeroMemory(cl, sizeof(CMDLIST));

				{
					char* p = line;

					cl->name = getToken(&p, '\t');
					cl->type = getTokenDword(&p, '\t');
					switch(cl->type)
					{
					case CMDTYPE_SYS:
						cl->param.syscmd = getTokenDword(&p, '\t');
						break;
					case CMDTYPE_EXE:
						cl->param.execmd.exe = getToken(&p, '\t');
						cl->param.execmd.param = getToken(&p, '\t');
						cl->param.execmd.workdir = getToken(&p, '\t');
						cl->param.execmd.nShowCmd = getTokenDword(&p, '\t');
						break;
					}
					cl->icon = getToken(&p, '\t');
				}
				cl->next = NULL;

				hfree(line);
			}

			CloseHandle(hf);
		}else
			MessageBox(NULL,	"cmdlist.txt܂BʂɂȂĂ悢łB\n"
								"\n"
								"\n"
								"name(tab)type(tab)param...(tab)icon\n"
								" type=1: system\n"
								"  param: systemcmd(0=quit 1=about)\n"
								" type=2: command\n"
								"  param: exefile(tab)parameter(tab)workdir(tab)1\n"
								"  Ō1SW_NORMAL̈ӖB\n",
								"",
								MB_ICONASTERISK | MB_OK);
	}

	while(GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	{
		PATHLIST* pl = rootpathlist;

		while(pl)
		{
			PATHLIST* next;

			hfree(pl->dir);
			next = pl->next;
			hfree(pl);
			pl = next;
		}
	}

	{
		CMDLIST* cl = rootcmdlist;

		while(cl)
		{
			CMDLIST* next;

			hfree(cl->name);	// ƂNULLłhfreełĂ邵[(O
			switch(cl->type)
			{
			case CMDTYPE_SYS:
				break;
			case CMDTYPE_EXE:
				hfree(cl->param.execmd.exe);
				hfree(cl->param.execmd.param);
				hfree(cl->param.execmd.workdir);
				break;
			}
			hfree(cl->icon);
			next = cl->next;
			hfree(cl);
			cl = next;
		}
	}

	DeleteObject(hInputFont);
	UnregisterHotKey(hMainWnd, 0);

	saveConfig();

	ExitProcess(msg.wParam);
}
